/* Top Level Object */
var Pull = function() {
	localize();

	//Props
	this.view = new PullView();

	//Init
	this.addViewListeners();
	this.addNBListeners();
	this.setToDefaults();
	var persist = this.readPersist();
	if(persist) {
		this.applyPersist(persist);
	} else {
		var flash = this.readFlash();
		if(flash) {
			this.applyFlash(flash);
		}
	}
	this.updateThumb();
	this.updateSizingOps();
	this.updateArrow();
};

Pull.prototype = {

	//Create
	setToDefaults: function() {
		this.view.layout({ //Default Layout
			width: $(window).width(),
			height: $(window).height(),
			isSettingsShown: false,
			orientation: 0,
		});
		this.view.setAutoSnap(true);
	},

	//Persist
	writePersist: function() {
		NB.persist.save('data', this.persistData());
	},

	persistData: function() {
		return {
			version: 1,
			rgb: this.view.rgb().slice(),
			title: $('#tabTitle').text(),
			text: $('#contentInput').val(),
			img: this.view.img(),
			orientation: this.view.orientation(),
			isSettingsShown: this.view.isSettingsShown(),
			isAutoSnap: this.view.isAutoSnap(),
		};
	},

	readPersist: function() {
		var str = NB.persist.load('data');
		if(!str) {
			return null;
		}
		var store = JSON.parse(str);
		if(!(store instanceof Object)) {
			return null;
		}
		var data = this.persistData(); //load defaults
		if(this.isValidRgb(store.rgb)) {
			data.rgb = store.rgb;
		}
		if(typeof store.title === 'string') {
			data.title = store.title;
		}
		if(typeof store.text === 'string') {
			data.text = store.text;
		}
		if(typeof store.img === 'string' && store.img !== '') {
			data.img = store.img;
		}
		var o = store.orientation
		if(typeof o === 'number' && o >= 0 && o < 4) {
			data.orientation = o;
		}
		data.isSettingsShown = Boolean(store.isSettingsShown);
		data.isAutoSnap = Boolean(store.isAutoSnap);
		return data;
	},

	applyPersist: function(data) {
		this.view.setRgb(data.rgb);
		this.view.setTitle(data.title);
		this.view.setContent(data.img, data.text);
		this.view.setAutoSnap(data.isAutoSnap);
		this.view.layout({
			width: $(window).width(),
			height: $(window).height(),
			isSettingsShown: data.isSettingsShown,
			orientation: data.orientation,
		});
	},

	isValidRgb: function(color) {
		if(!color) {
			return false;
		}
		for(var i = 0; i < 3; i++) {
			var x = color[i];
			if(!(typeof x === 'number') || x < 0 || x > 255) {
				return false;
			}
		}
		return true;
	},

	//Flash
	readFlash: function() {
		var text = '';
		var img = '';
		var title = '';
		var orientation = 0;
		try {
			var perData = new XmlDocument(NB.persist.load('updateData'));
			var textProps = this.getFlashChild(perData, 'Text');
			text = textProps.childNamed('InputText').val;
			img = NB.persist.loadFile('my_mc0');
			title  = textProps.childNamed('tabText').val;

			//Orientation
			var boolProps = this.getFlashChild(perData, 'Boolean');
			var tabLocation = Number(boolProps.childNamed('tabLocation').val);
			switch(tabLocation) {
				case 0:
					orientation = 2;
					break;
				case 1:
					orientation = 3;
					break;
				case 2:
					orientation = 0;
					break;
				case 3:
					orientation = 1;
					break;
			}
		} catch(e) {
			return null;
		}

		//Snaps 
		//Note that for flash widgets the layout is a bit strange. There are tabs on both sides but one is not shown.
		//This is why the snapContent calcs take into account tabSize as well.
		var host = NB.getHostObject();
		var x = host.x;
		var y = host.y;
		var x2 = x + host.width;
		var y2 = y + host.height;
		var snapContent = false;
		var snapTab = false;
		var thresh = 10;
		var tabSize = 35 / 450 * host.width;
		switch(orientation) {
			case 0:
				snapContent = x + tabSize > -thresh && x + tabSize < thresh;
				snapTab = x2 - tabSize > -thresh && x2 - tabSize < thresh;
				break;
			case 1:
				snapContent = y + tabSize > -thresh && y + tabSize < thresh;
				snapTab = y2 - tabSize > -thresh && y2 - tabSize < thresh;
				break;
			case 2:
				var pageWidth = NB.page.width;
				snapContent = x2 - tabSize > pageWidth - thresh && x2 - tabSize < pageWidth + thresh;
				snapTab = x + tabSize > pageWidth - thresh && x + tabSize < pageWidth + thresh;
				break;
			case 3:
				var pageHeight = NB.page.height;
				snapContent = y2 - tabSize > pageHeight - thresh && y2 - tabSize < pageHeight + thresh;
				snapTab = y + tabSize > pageHeight - thresh && y + tabSize < pageHeight + thresh;
				break;
		}

		return {
			title: title,
			text: text,
			img: img,
			orientation: orientation,
			snapContent: snapContent,
			snapTab: snapTab,
		};
	},

	getFlashChild: function(persistenceData, id) {
		var kids = persistenceData.children;
		for(var i = 0; i < kids.length; i++) {
			var k = kids[i];
			if(k.name === 'PersistenceProperties' && k.attr.id === id) {
				return k;
			}
		}
		return null;
	},

	applyFlash: function(data) {
		var min = this.minSizesFor(data.orientation, false);
		var width = Math.max($(window).width(), min.width);
		var height = Math.max($(window).height(), min.height);
		var host = NB.getHostObject();
		this.setNBRect(host.x, host.y, width + 2 * this.annMarg(), height + 2 * this.annMarg());
		this.view.setAutoSnap(false);
		if(!this.isDefaultFlashText(data.text)) {
			$('#contentInput').val(data.text); //Only apply text if it not the old placeholder
		}
		
		if(data.title) {
			this.view.setTitle(data.title);
		}
		this.view.layout({
			width: width,
			height: height,
			isSettingsShown: false,
			orientation: data.orientation,
		});
		if(data.img) {
			new ImgUrlConverter(data.img, this.onImgConvert.bind(this));
		}
		if(data.snapContent) {
			this.doSnap(data.orientation, true);
		} else if(data.snapTab) {
			this.doSnap(data.orientation, false);
		}
		this.writePersist();	
	},

	isDefaultFlashText: function(text) {
		//Get Alphabetical chars
		var alphaChars = '';
		for(var i = 0; i < text.length; i++) {
			var charCode = text.charCodeAt(i);
			if((charCode >= 65 && charCode <= 90) || (charCode >= 97 && charCode <= 122)) {
				alphaChars += text[i];
			}
		}

		//Check
		return alphaChars === 'EntertexthereUsethepintosettheendpositionandclickthearrowtogobacktothatlocation';
	},

	onImgConvert: function(converter, newUrl) {
		this.view.setImg(newUrl);
		this.writePersist();
	},

	//NB
	addNBListeners: function() {
		NB.addObserver('dndDropEvent', this.dndDropEvent.bind(this));
	},

	updateThumb: function() {
		// add delay to give nb some time to update its internal bitmaps
		setTimeout(function() {
			NB.page.updateThumbnail();
		}, 100);
	},

	dndDropEvent: function(event) {
		var data = event.eventData;
		var hostGuid = NB.getHostObject().guid;
		for(var guid in data) {
			if(guid === hostGuid) {
				this.onAnnDrop(event);
			} else {
				this.dropCheck(event, guid);
			}
		}
		var guid = Object.keys(data)[0];
		ann = NB.getObject(guid);
	},

	//DND
	dropCheck: function(event, annGuid) {
		try {
			var host = NB.getHostObject();
			var ann = NB.getObject(annGuid);
			var contact = event.eventData[annGuid].contact;
			var marg = this.annMarg();
			var x = contact.x - host.x - marg;
			var y = contact.y - host.y - marg;
			if(this.pointOverlapsElement(x, y, $('#contentInput'))) {
				if(ann.type === 'text') {
					new TextDrop(this, annGuid, false);
				} else {
					new ImgDrop(this, annGuid);
				}
			} else if(this.pointOverlapsElement(x, y, $('#titleInput')) && ann.type === 'text') {
				new TextDrop(this, annGuid, true);
			}
		} catch(e) {}
	},

	pointOverlapsElement: function(x, y, jq) {
		var offset = jq.offset();
		var left = offset.left;
		var top = offset.top;
		var right = left + jq.width();
		var bottom = top + jq.height();
		return x >= left && x <= right && y >= top && y <= bottom;
	},

	//Sizing
	annMarg: function() {
		/* between annotation bounding box and widget content */

		// _annMarg is a "static" variable being initialized once
		if (typeof this._annMarg === "undefined")
		{
			var host = NB.getHostObject();
			// annotation "margin" added in NB next
			this._annMarg = typeof host.margin == "undefined" ? 9 : host.margin;
		}

		return this._annMarg;
	},

	setNBRect: function(x, y, w, h) {
		NB.undomanager.setAutoRecord(false);
		NB.batch(function() {
			var host = NB.getHostObject();
			host.width = w;
			host.height = h;
			host.x = x;
			host.y = y;
		});
		NB.undomanager.setAutoRecord(true);
	},

	offsetNBRectSafe: function(dx, dy, dw, dh) {
		/* Makes sure annotation is not hidden. */
		var host = NB.getHostObject();
		var xNew = host.x + dx;
		var yNew = host.y + dy;
		var wNew = host.width + dw;
		var hNew = host.height + dh;
		var marg = this.annMarg();
		var xMax = NB.page.width + marg - wNew;
		var yMax = NB.page.height + marg - hNew;
		
		if(xNew < -marg) {
			xNew = -marg;
		} else if(xNew > xMax) {
			xNew = xMax;
		}
		if(yNew < -marg) {
			yNew = -marg;
		} else if(yNew > yMax) {
			yNew = yMax;
		}
		this.setNBRect(xNew, yNew, wNew, hNew);
	},

	setNBRect2: function(x, y, dw, dh) {
		var host = NB.getHostObject();
		this.setNBRect(x, y, host.width + dw, host.height + dh);
	},

	updateSizingOps: function() {
		var host = NB.getHostObject();
		var min = this.minSizesFor(this.view.orientation(), this.view.isSettingsShown());
		host.setWidgetMinSize(min.width, min.height);
	},

	minSizesFor: function(orientation, settingsShown) {
		var marg = this.annMarg();
		var width = 160 + 2 * marg;
		var height = 180 + 2 * marg;
		if(orientation === 0 || orientation === 2) {
			width += this.view.tabSize();
		} else {
			height += this.view.tabSize();
		}
		if(settingsShown) {
			width += $('#settings').width();
		}
		return {
			width: width,
			height: height,
		};
	},

	onResize: function() {
		this._resize();
		setTimeout(this._resize.bind(this), 10); //Sometimes sent improper coords so update later on if need be.
	},

	_resize: function() {
		var w = $(window).width();
		var h = $(window).height();
		var v = $('#pullView');
		if(v.width() !== w || v.height() !== h) {
			this.view.setSize(w, h);
			this.updateArrow();
			this.writePersist();
		}
	},

	//Moves
	closestSide: function(event) {
		/* Return closest side.  0 == left, 1 == top, 2 == right, 3 == bottom */
		var host = NB.getHostObject();
		var contact = event.eventData[host.guid].contact;
		var page = NB.page;
		var x = contact.x - GetPageLeft(page);
		var y = contact.y - GetPageTop(page);
		var pageWidth = page.width;
		var pageHeight = page.height;
		
		var side = 0;
		var min = x;
		
		//Check Top
		if(y < min) {
			min = y;
			side = 1;
		}

		//Check Right
		if(pageWidth - x < min) {
			min = pageWidth - x;
			side = 2;
		}

		//Check Bottom
		if(pageHeight - y < min) {
			min = pageHeight - y;
			side = 3;
		}
		return side;
	},

	updateArrow: function() {
		var pointOut = this.isWidgetOutOfPageBounds();
		this.view.setArrowPointOut(pointOut);
	},

	doSnap: function(orientation, showContent) {
		if(orientation !== this.view.orientation()) {
			var x = new FadedAction({
				onMid: this._doSnapFaded.bind(this),
				onEnd: this.onFadedEnd.bind(this),
				props: {
					orientation: orientation,
					showContent: showContent,
				},
			});
			x.start();
		} else {
			this._doSnap(orientation, showContent);
		}
	},

	_doSnapFaded: function(action) {
		this._doSnap(action.props.orientation, action.props.showContent);
	},

	_doSnap: function(orientation, showContent) {
		var v = $('#pullView');
		var host = NB.getHostObject();
		var page = NB.page;
		var pageWidth = page.width;
		var pageHeight = page.height;
		var marg = this.annMarg();
		var tabSize = this.view.tabSize();
		var dif = this.view.switchOrientation(orientation);
		var newW = host.width + dif.dw;
		var newH = host.height + dif.dh;
		var newX = host.x;
		var newY = host.y;
		switch(orientation) {
			case 0:
				newX = showContent ? -marg : -newW + tabSize + marg;
				newX += GetPageLeft(page);
				break;
			case 1:
				newY = showContent ? -marg : -newH + tabSize + marg;
				newY += GetPageTop(page);
				break;
			case 2:
				newX = showContent ? pageWidth - marg - v.width() : pageWidth - tabSize - marg;
				newX += GetPageLeft(page);
				break;
			case 3:
				newY = showContent ? pageHeight - marg - v.height() : pageHeight - tabSize - marg;
				newY += GetPageTop(page);
				break;
		}
		this.setNBRect(newX, newY, newW, newH);
		this.updateArrow();
		this.updateSizingOps();
		this.writePersist();
	},

	doMove: function(orientation, x, y) {
		if(orientation !== this.view.orientation()) {
			var x = new FadedAction({
				onMid: this._doMoveFaded.bind(this),
				onEnd: this.onFadedEnd.bind(this),
				props: {
					orientation: orientation,
					x: x,
					y: y,
				},
			});
			x.start();
		} else {
			this._doMove(orientation, x, y);
		}
	},

	_doMoveFaded: function(action) {
		this._doMove(action.props.orientation, action.props.x, action.props.y);
	},

	_doMove: function(orientation, x, y) {
		var host = NB.getHostObject();
		var dif = this.view.switchOrientation(orientation);
		var newW = host.width + dif.dw;
		var newH = host.height + dif.dh;
		this.setNBRect(x, y, newW, newH);
		this.updateArrow();
		this.updateSizingOps();
		this.writePersist();
	},

	onAnnDrop: function(event) {
		var host = NB.getHostObject();
		if(host.rotation === 0 && this.view.isAutoSnap()) {
			var marg = this.annMarg();
			var page = NB.page;
			var x = host.x + marg - GetPageLeft(page);
			var y = host.y + marg - GetPageTop(page);
			var w = host.width - 2 * marg;
			var h = host.height - 2 * marg;
			var x2 = x + w;
			var y2 = y + h;
			var wPage = page.width;
			var hPage = page.height;
			
			var tabSize = this.view.tabSize();
			var threshold = 30;
			

			//Determine new orientation
			var newOrientation = this.view.orientation();
			switch(newOrientation) {
				case 0:
					if(x > threshold) {
						newOrientation = this.closestSide(event);
					}
					break;
				case 1:
					if(y > threshold) {
						newOrientation = this.closestSide(event);
					}
					break;
				case 2:
					if(wPage - x2 > threshold) {
						newOrientation = this.closestSide(event);
					}
					break;
				case 3:
					if(hPage - y2 > threshold) {
						newOrientation = this.closestSide(event);
					}
					break;
			}

			//Determine shouldSnap and showContent
			var showContent = false;
			var shouldSnap = false;
			switch(newOrientation) {
				case 0:
					showContent = x > -w / 2;
					shouldSnap = x < threshold;
					break;
				case 1:
					showContent = y > -h / 2;
					shouldSnap = y < threshold;
					break;
				case 2:
					showContent = x < wPage - w / 2;
					shouldSnap = wPage - x2 < threshold;
					break;
				case 3:
					showContent = y < hPage - h / 2;
					shouldSnap = hPage - y2 < threshold;
					break;
			}
			if(newOrientation === this.view.orientation()) {
				switch(newOrientation) {
					case 0:
						shouldSnap = Math.abs(x) < threshold || (x2 - tabSize < threshold);
						break;
					case 1:
						shouldSnap = Math.abs(y) < threshold || (y2 - tabSize < threshold);
						break;
					case 2:
						shouldSnap = Math.abs(x2 - wPage) < threshold || (x + tabSize > wPage - threshold);
						break;
					case 3:
						shouldSnap = Math.abs(y2 - hPage) < threshold || (y + tabSize > hPage - threshold);
						break;
				}
				
			}

			//Make call
			if(shouldSnap) {
				new Move(this, newOrientation, showContent);
			}
		}
		this.updateArrow();
	},

	//Toggle Settings
	toggleSettings: function() {
		var show = !this.view.isSettingsShown();
		new ToggleSettings(this, show);
	},

	setSettingsShown: function(settingsShown) {
		if(this.view.orientation() == 2) {
			var x = new FadedAction({
				onMid: this._setSettingsShownFaded.bind(this),
				onEnd: this.onFadedEnd.bind(this),
				props: {
					settingsShown: settingsShown,
				},
			});
			x.start();
		} else {
			this._setSettingsShown(settingsShown);
		}
	},

	_setSettingsShownFaded: function(action) {
		this._setSettingsShown(action.props.settingsShown);
	},

	_setSettingsShown: function(settingsShown) {
		var old = $('#pullView').width();
		this.view.setIsSettingsShown(settingsShown);
		var dw = $('#pullView').width() - old;
		var dx = this.view.orientation() === 2 ? -dw : 0;
		this.offsetNBRectSafe(dx, 0, dw, 0);
		this.updateSizingOps();
		this.writePersist();
	},

	onFadedEnd: function() {
		this.updateThumb();
	},

	//Events
	addViewListeners: function() {
		this.addColorListeners();
		$('.tabPos').click(this.onTabPosClick.bind(this));
		$('#arrowOuter').click(this.onArrow.bind(this));
		$('#gear').click(this.onGear.bind(this));
		$('#autoSnapBox').click(this.onAutoSnapClick.bind(this));
		$('#contentImg').mousedown(this.onContentImgMouseDown.bind(this));
		$('#titleInput').on('input propertychange paste', this.onTitleInput.bind(this));
		$('#titleInput').keydown(this.stopProp);
		$('#contentInput').keydown(this.stopProp);
		$('#contentInput').on('input propertychange paste', this.onContentInput.bind(this));
		$('#pullView').mousedown(this.onMainMouseDown.bind(this));
		$(document).keydown(this.onKey.bind(this));
		$(document).focusout(this.onFocusOut.bind(this));
		$(window).resize(this.onResize.bind(this));
	},

	//Content Events
	onTitleInput: function(event) {
		$('#tabTitle').text($('#titleInput').val());
		this.view.layoutTitle();
		this.view.layoutTitleInput();
		this.writePersist();
		this.updateThumb();
	},

	onContentInput: function(event) {
		this.writePersist();
		this.updateThumb();
	},

	onArrow: function() {
		var showContent = this.isWidgetOutOfPageBounds();
		new Move(this, this.view.orientation(), showContent);
	},

	onContentImgMouseDown: function(event) {
		event.stopPropagation();
		this.view.hideContextViews();
		this.view.setImgSelected(true);
		this.updateThumb();
	},

	//Settings Events
	onTabPosClick: function(event) {
		var orient = this.view.btnToOrientation(event.currentTarget);
		new Move(this, orient, true);
	},

	onGear: function() {
		this.toggleSettings();
	},

	onAutoSnapClick: function() {
		this.view.toggleAutoSnap();
		this.writePersist();
	},
	
	//Color Events
	addColorListeners: function() {
		var v = this.view.picker;
		v.onHexClick = this.onHexClick.bind(this);
		v.onWheelUp = this.onWheelUp.bind(this);
		v.onDrag = this.onWheelDrag.bind(this);
		v.onXClick = this.onWheelXClick.bind(this);
		$('#pickerWrap').mousedown(this.onPickerWrapMouseDown.bind(this));
		$('#colorBox').click(this.onColorBox.bind(this))
			.mousedown(this.onColorBoxMouseDown.bind(this));
	},

	onColorBox: function(event) {
		this.view.togglePicker();
		event.stopPropagation();
		this.updateThumb();
	},

	onColorBoxMouseDown: function(event) {
		event.stopPropagation();
		this.view.setImgSelected(false);
	},

	onHexClick: function(picker, newRgb) {
		this.view.setRgb(newRgb, true);
		this.writePersist();
		this.updateThumb();
	},

	onWheelUp: function(picker, newRgb) {
		this.view.setRgb(newRgb, true);
		this.writePersist();
		this.updateThumb();
	},

	onWheelDrag: function(picker, newRgb) {
		this.view.setRgb(newRgb, true);
	},

	onWheelXClick: function(picker) {
		$('#pickerWrap').hide();
		this.updateThumb();
	},

	onPickerWrapMouseDown: function(event) {
		event.stopPropagation();
		this.view.setImgSelected(false);
	},

	//Other Events
	onKey: function(event) {
		var k = event.which;
		if((k === 8 || k === 46) && this.view.isImgSelected()) {
			this.view.removeImg(false);
			this.writePersist();
			this.updateThumb();
		}
	},

	onFocusOut: function() {
		this.view.hideContextViews();
	},

	onMainMouseDown: function() {
		this.view.hideContextViews();
	},

	stopProp: function(event) {
		event.stopPropagation();
	},

	//Other
	nextTo: function(x, y, threshold) {
		/* returns true if x and y are less than threshold apart */
		return Math.abs(x - y) < threshold;
	},
	
	isWidgetOutOfPageBounds: function isWidgetOutOfPageBounds()
	{
		var host = NB.getHostObject();
		var page = NB.page;
		var x = host.x - GetPageLeft(page);
		var y = host.y - GetPageTop(page);
		var w = $('#pullView').width();
		var h = $('#pullView').height();
		var outOfPage = false;
		var marg = this.annMarg();
		switch(this.view.orientation()) {
			case 0:
				outOfPage = x + marg < 0;
				break;
			case 1:
				outOfPage = y + marg < 0;
				break;
			case 2:
				outOfPage = x + w - marg > page.width;
				break;
			case 3:
				outOfPage = y + h - marg > page.height;
				break;
		}
		return outOfPage;
	}

};

NB.ready(function() {
	if (window.requiresNewerNBSDKVersionError) {
        var container = window.document.createElement("div");
        container.setAttribute("id", "requiresNewerNBSDKVersionErrorContainer");

        var p1 = window.document.createElement("p");
        p1.textContent = NB.localize("This activity requires a later version of SMART Notebook software.");
        container.appendChild(p1);

        var p2 = window.document.createElement("p");
        p2.textContent = NB.localize("Make sure you're using Notebook software version %d or later.").replace("%d", "16");
        container.appendChild(p2);

        document.body.appendChild(container);
        document.body.style.backgroundColor = "lightgray";

        function adjustHeights(id) {
            var elem = document.getElementById(id);
            var fontstep = 2;
            if (elem.offsetHeight > document.documentElement.clientHeight) {
                elem.style.fontSize = ((window.getComputedStyle(elem).fontSize.substr(0,2)-fontstep).toString() + 'px');
                elem.style.lineHeight = (window.getComputedStyle(elem).fontSize.substr(0,2).toString() + 'px');
                adjustHeights(id);
            }
        }

        adjustHeights("requiresNewerNBSDKVersionErrorContainer");

        return;
    }

	x = new Pull();
});

function GetPageLeft(page)
{
	// NB.page.left added in NB next
	return typeof page.left === "undefined" ? 0 : page.left;
}

function GetPageTop(page)
{
	// NB.page.top added in NB next
	return typeof page.top === "undefined" ? 0 : page.top;
}
